# -*- coding: utf-8 -*-

# penalty parameter to enforce essential boundary conditions
# using a Courant penalty function
DEFAULT_DIRICHLET_PENALTY = 1E12

# a ratio of the penalty factor applied for the stiffness and
# mass matrices for the bipenalty method
DEFAULT_MASS_PENALTY_RATIO = 1E5
